@@ -4,7 +4,7 @@
 # you may not use this file except in compliance with the License.
 # You may obtain a copy of the License at
 #
-# http://www.apache.org/licenses/LICENSE-2.0
+# https://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS,
@@ -21,6 +21,9 @@ import getpass
 import sys
 import argparse
 import warnings
+import collections
+
+from requests.exceptions import HTTPError
 
 try:
     import configparser
@@ -32,11 +35,19 @@ try:
 except ImportError:
     from urllib.parse import urlparse, urlunparse
 
+try:
+    import keyring  # noqa
+except ImportError:
+    pass
+
+from twine import exceptions
+
 # Shim for raw_input in python3
 if sys.version_info > (3,):
     input_func = input
 else:
-    input_func = raw_input
+    # Ignore "undefined name" for flake8/python3
+    input_func = raw_input  # noqa: F821
 
 
 DEFAULT_REPOSITORY = "https://upload.pypi.org/legacy/"
@@ -44,74 +55,68 @@ TEST_REPOSITORY = "https://test.pypi.org/legacy/"
 
 
 def get_config(path="~/.pypirc"):
+    # even if the config file does not exist, set up the parser
+    # variable to reduce the number of if/else statements
+    parser = configparser.RawConfigParser()
+
+    # this list will only be used if index-servers
+    # is not defined in the config file
+    index_servers = ["pypi", "testpypi"]
+
+    # default configuration for each repository
+    defaults = {"username": None, "password": None}
+
     # Expand user strings in the path
     path = os.path.expanduser(path)
 
-    if not os.path.isfile(path):
-        return {"pypi": {"repository": DEFAULT_REPOSITORY,
-                         "username": None,
-                         "password": None
-                         },
-                "pypitest": {"repository": TEST_REPOSITORY,
-                             "username": None,
-                             "password": None
-                             },
-                }
-
     # Parse the rc file
-    parser = configparser.RawConfigParser()
-    parser.read(path)
+    if os.path.isfile(path):
+        parser.read(path)
 
-    # Get a list of repositories from the config file
-    if (parser.has_section("distutils") and
-            parser.has_option("distutils", "index-servers")):
-        repositories = parser.get("distutils", "index-servers").split()
-    else:
-        repositories = ["pypi"]
+        # Get a list of index_servers from the config file
+        # format: https://docs.python.org/3/distutils/packageindex.html#pypirc
+        if parser.has_option("distutils", "index-servers"):
+            index_servers = parser.get("distutils", "index-servers").split()
 
-    config = {}
-
-    defaults = {"username": None, "password": None}
-    if parser.has_section("server-login"):
         for key in ["username", "password"]:
             if parser.has_option("server-login", key):
                 defaults[key] = parser.get("server-login", key)
 
-    for repository in repositories:
-        # Skip this repository if it doesn't exist in the config file
-        if not parser.has_section(repository):
-            continue
+    config = collections.defaultdict(lambda: defaults.copy())
 
-        # Mandatory configuration and defaults
-        config[repository] = {
-            "repository": DEFAULT_REPOSITORY,
-            "username": None,
-            "password": None,
-        }
+    # don't require users to manually configure URLs for these repositories
+    config["pypi"]["repository"] = DEFAULT_REPOSITORY
+    if "testpypi" in index_servers:
+        config["testpypi"]["repository"] = TEST_REPOSITORY
 
-        # Optional configuration values
+    # optional configuration values for individual repositories
+    for repository in index_servers:
         for key in [
             "username", "repository", "password",
             "ca_cert", "client_cert",
         ]:
             if parser.has_option(repository, key):
                 config[repository][key] = parser.get(repository, key)
-            elif defaults.get(key):
-                config[repository][key] = defaults[key]
 
-    return config
+    # convert the defaultdict to a regular dict at this point
+    # to prevent surprising behavior later on
+    return dict(config)
 
 
 def get_repository_from_config(config_file, repository, repository_url=None):
-    # Get our config from the .pypirc file
+    # Get our config from, if provided, command-line values for the
+    # repository name and URL, or the .pypirc file
     if repository_url and "://" in repository_url:
-        # assume that the repository is actually an URL and just sent
-        # them a dummy with the repo set
+        # prefer CLI `repository_url` over `repository` or .pypirc
         return {
             "repository": repository_url,
             "username": None,
             "password": None,
         }
+    if repository_url and "://" not in repository_url:
+        raise exceptions.UnreachableRepositoryURLDetected(
+            "Repository URL {} has no protocol. Please add "
+            "'https://'. \n".format(repository_url))
     try:
         return get_config(config_file)[repository]
     except KeyError:
@@ -125,11 +130,11 @@ def get_repository_from_config(config_file, repository, repository_url=None):
             repo=repository,
             cfg=config_file
         )
-        raise KeyError(msg)
+        raise exceptions.InvalidConfiguration(msg)
 
 
-_HOSTNAMES = set(["pypi.python.org", "testpypi.python.org", "upload.pypi.org",
-                  "test.pypi.org"])
+_HOSTNAMES = {"pypi.python.org", "testpypi.python.org", "upload.pypi.org",
+              "test.pypi.org"}
 
 
 def normalize_repository_url(url):
@@ -139,16 +144,32 @@ def normalize_repository_url(url):
     return urlunparse(parsed)
 
 
-def check_status_code(response):
-    if (response.status_code == 500 and
+def check_status_code(response, verbose):
+    """
+    Shouldn't happen, thanks to the UploadToDeprecatedPyPIDetected
+    exception, but this is in case that breaks and it does.
+    """
+    if (response.status_code == 410 and
             response.url.startswith(("https://pypi.python.org",
                                      "https://testpypi.python.org"))):
-        print("It appears you're uploading to pypi.python.org (or testpypi) "
-              "you've received a 500 error response. PyPI is being phased "
-              "out for pypi.org. Try using https://upload.pypi.org/legacy/ "
+        print("It appears you're uploading to pypi.python.org (or "
+              "testpypi.python.org). You've received a 410 error response. "
+              "Uploading to those sites is deprecated. The new sites are "
+              "pypi.org and test.pypi.org. Try using "
+              "https://upload.pypi.org/legacy/ "
               "(or https://test.pypi.org/legacy/) to upload your packages "
-              "instead. These are the default URLs for Twine now.")
-    response.raise_for_status()
+              "instead. These are the default URLs for Twine now. More at "
+              "https://packaging.python.org/guides/migrating-to-pypi-org/ ")
+    try:
+        response.raise_for_status()
+    except HTTPError as err:
+        if response.text:
+            if verbose:
+                print('Content received from server:\n{}'.format(
+                    response.text))
+            else:
+                print('NOTE: Try --verbose to see response content.')
+        raise err
 
 
 def get_userpass_value(cli_value, config, key, prompt_strategy=None):
@@ -174,7 +195,7 @@ def get_userpass_value(cli_value, config, key, prompt_strategy=None):
     """
     if cli_value is not None:
         return cli_value
-    elif config.get(key):
+    elif config.get(key) is not None:
         return config[key]
     elif prompt_strategy:
         return prompt_strategy()
@@ -182,6 +203,23 @@ def get_userpass_value(cli_value, config, key, prompt_strategy=None):
         return None
 
 
+def get_username_from_keyring(system):
+    if 'keyring' not in sys.modules:
+        return
+
+    try:
+        getter = sys.modules['keyring'].get_credential
+    except AttributeError:
+        return None
+
+    try:
+        creds = getter(system, None)
+        if creds:
+            return creds.username
+    except Exception as exc:
+        warnings.warn(str(exc))
+
+
 def password_prompt(prompt_text):  # Always expects unicode for our own sanity
     prompt = prompt_text
     # Workaround for https://github.com/pypa/twine/issues/116
@@ -191,17 +229,22 @@ def password_prompt(prompt_text):  # Always expects unicode for our own sanity
 
 
 def get_password_from_keyring(system, username):
-    try:
-        import keyring
-    except ImportError:
+    if 'keyring' not in sys.modules:
         return
 
     try:
-        return keyring.get_password(system, username)
+        return sys.modules['keyring'].get_password(system, username)
     except Exception as exc:
         warnings.warn(str(exc))
 
 
+def username_from_keyring_or_prompt(system):
+    return (
+        get_username_from_keyring(system)
+        or input_func('Enter your username: ')
+    )
+
+
 def password_from_keyring_or_prompt(system, username):
     return (
         get_password_from_keyring(system, username)
@@ -209,11 +252,18 @@ def password_from_keyring_or_prompt(system, username):
     )
 
 
-get_username = functools.partial(
-    get_userpass_value,
-    key='username',
-    prompt_strategy=functools.partial(input_func, 'Enter your username: '),
-)
+def get_username(system, cli_value, config):
+    return get_userpass_value(
+        cli_value,
+        config,
+        key='username',
+        prompt_strategy=functools.partial(
+            username_from_keyring_or_prompt,
+            system,
+        ),
+    )
+
+
 get_cacert = functools.partial(
     get_userpass_value,
     key='ca_cert',
@@ -253,3 +303,31 @@ def get_password(system, username, cli_value, config):
             username,
         ),
     )
+
+
+def no_positional(allow_self=False):
+    """A decorator that doesn't allow for positional arguments.
+
+    :param bool allow_self:
+        Whether to allow ``self`` as a positional argument.
+    """
+    def reject_positional_args(function):
+        @functools.wraps(function)
+        def wrapper(*args, **kwargs):
+            allowed_positional_args = 0
+            if allow_self:
+                allowed_positional_args = 1
+            received_positional_args = len(args)
+            if received_positional_args > allowed_positional_args:
+                function_name = function.__name__
+                verb = 'were' if received_positional_args > 1 else 'was'
+                raise TypeError(('{}() takes {} positional arguments but {} '
+                                 '{} given').format(
+                                     function_name,
+                                     allowed_positional_args,
+                                     received_positional_args,
+                                     verb,
+                                ))
+            return function(*args, **kwargs)
+        return wrapper
+    return reject_positional_args
